/* Tutorial on Reverse Search */
/* Section 4: C Implementation for Permutations */
/* Solution to Ex 2.2 topological sorts */

/* Input: enter number n of vertices and m edges          */
/*        followed by m pairs i j with 1 <=i < j <= n     */ 
/*        this implies that 1,2,...,n is a topsort        */


/* This code is derived from topsorts.c and is standalone single thread program */
/* It is further modified in btopsorts.c to allow multithreading                */

/* gcc -O3 -o atopsorts atopsorts.c                                             */

/* Usage  atopsorts < file [-countonly] [-maxnodes n] [-maxd d]                 */

/* pm6 is a sample input file that generates 15 permutations                    */

/* Changes include:
   1. internal data grouped in bts_data structure
   2. initialization of bts_data from the input file via bts_init
   3. max_nodes parameter implemented
   4. max_depth parameter implemented
   5. unexplored subtrees created by 3,4 are returned with *unexplored flag
   6. reverse search initiated from any subtree root 
   7. command line arguments allowed
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

#define TRUE 1
#define FALSE 0

typedef int perm[100];

struct bts_data;
typedef struct bts_data bts_data;


/* bts_data is used to hold input data  */
/* except where mentioned               */

struct bts_data {
        int n;                 /* number of nodes and edges in graph */
        int m;
        int A[100][100];       /* the graph */
        int countonly;         /* TRUE means nodes are counted but not output */
};

/* bts_init is executed at startup to process command line and build bts_data */

bts_data *bts_init(int argc, char **argv)
{
  int i,j,k,m,n;

  bts_data *b = malloc(sizeof(bts_data));

  b->countonly=FALSE;

/* process command line arguments         */

        for (i=1; i < argc; i++)
          if (strcmp(argv[i],"-countonly") == 0)
             b->countonly=TRUE;

/* read input file and build bts_data     */

  scanf ("%d %d", &b->n,&b->m);
  n=b->n;
  m=b->m;
  for (i=1; i<=n; i++)
     for (j=1; j<=n; j++)
       b->A[i][j]=0;

  for (k=1; k<=m; k++)
    {
     scanf("%d %d",&i,&j);
     if (i >= j)
       {  printf ("\nEdge must have form i<j");
          return FALSE;
       }
     b->A[i][j]=1;
    }

        return b;
}

void put_output (bts_data *b, perm v, int depth, int unexplored)
{
  int i;
  int n=b->n;
  if(!b->countonly)
    {
      for (i = 1; i <= n; i++)
           printf (" %d", v[i]);
      printf (" d=%d ",depth);
      if(unexplored)
           printf (" *unexplored");
      printf ("\n");
     }
}

void copy (bts_data *b, perm w, perm v)
{
  int i;
  int n = b->n;

  for (i = 1; i <= n; i++)
    w[i] = v[i];
}

int equal (bts_data *b, perm w, perm v)
{
  int i, n=b->n;
  for (i = 1; i <= n; i++)
    if (w[i] != v[i])
      return FALSE;
  return TRUE;
}


void swap (perm v, int j)
{
  int t;
  t = v[j];
  v[j] = v[j + 1];
  v[j + 1] = t;
}

void Adj (perm v, int j)	/* adjacency oracle */
{
  swap (v, j);
}

void f (bts_data *b, perm v)			/* local search function */
{
  int i=1, n=b->n;
  while ( i < n && v[i] <= v[i + 1] ) i++;
  if (i < n)
    swap (v, i);
}

int reverse (bts_data *b, perm v, int j)
{
  if(b->A[v[j]][v[j+1]]==1) return (FALSE);  /* v[j]v[j+1] is an edge */
  perm w;
  copy (b, w, v);
  Adj (w, j);
  f (b, w);
  return equal (b, v, w);
}

int backtrack (bts_data *b, perm v)
{
  perm child;
  int j = 0;
  int found = FALSE;
  copy (b, child, v);
  f (b, v);			/* v is parent of child */
  do{
      j++;
      Adj (v, j);               /* v replaced by jth child */
      found = equal (b, child, v);
      swap (v, j);		/* restore v */
     }  while (!found);
  return j;
}

long bts(bts_data *b, perm v, long max_depth, long max_nodes)

/* budgeted reverse search code for topsorts */

{
  int j=0, depth=0, unexplored;
  long count=0;
  int n = b->n;
  int maxdeg=n-1;;

  do
  {
      unexplored = FALSE;
      while (j < maxdeg && !(unexplored) )
      {
         j++;
         if (reverse (b, v, j))
          {                        /* forward step */
             Adj (v, j);
             (depth)++;
             count++;
             if (count >= max_nodes || (depth) == max_depth)
                  unexplored=TRUE;
	     put_output(b, v,depth,unexplored);
             j = 0;
          }
      }   /* end while */

     if(depth > 0)
          {                        /* backtrack step */
             j = backtrack (b, v);
             (depth)--;
          }
 } while( (depth) > 0 || j < maxdeg);
  return count;
}


int main(int argc, char **argv) 
/* generate topsorts with budgeting and bts_data structure */
{
  bts_data *b;
  perm v;
  long count;
  int i,n;
  long max_nodes=LONG_MAX, max_depth=LONG_MAX;

  printf("\n*atopsorts 2016.6.3:");
  printf("\n");

  b=bts_init(argc,argv);
  n=b->n;

  for (i = 1; i <= n; i++)
    v[i] = i;

  max_depth=LONG_MAX;
  max_nodes=LONG_MAX;   

  for (i=1; i < argc; i++)
    {
          if (strcmp(argv[i],"-maxd") == 0)
                max_depth = atoi(argv[i+1]);
          if (strcmp(argv[i],"-maxnodes") == 0)
                 max_nodes = atoi(argv[i+1]);
    }


/* the tree search is now initiated from root */

  put_output(b,v,0,0);     /* first output */
  count = bts(b,v,max_depth,max_nodes);  

/* final totals */
  printf ("\nnumber of permutations=%ld\n", count+1); /* count the root */

  free(b);
  return 0;
}
