/* Tutorial on Reverse Search          */
/* Section 6: Tree generation  : Ex 6.3*/
/* Modified for multithreading         */

/* Edge list version                   */

/* input: number N of vertices  and number m of edges in underlying graph */
/*        followed by an edge list ij                                     */
/*        vertices labelled 0,1,...,N-1=n                                 */

/* This code is derived from tree.c and is standalone single thread program */
/* that is further modified to allow multithreading                         */

/* gcc -O3 -o atree atree.c                                                 */

/* Usage  atree < file [-countonly] [-maxnodes n] [-maxd d] [-prune 0/1]    */

/* k23 is a sample input file that generates 12 trees                       */

/* Initial changes include:
   1. internal data grouped in bts_data structure
   2. initialization of bts_data from the input file via bts_init
   3. max_nodes parameter implemented
   4. max_depth parameter implemented
   5. unexplored subtrees created by 3,4 are returned with *unexplored flag
   6. reverse search initiated from any subtree root
   7. command line arguments allowed
   8. -prune 0 does not print leaves with *unexplored 
   9. -prune 1 does not return leaves and follows single child edges       
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define TRUE 1
#define FALSE 0
#define MAXN 100


typedef long tree[100];  /* nodes from 0..n, tree[i] is parent of i */

struct bts_data;
typedef struct bts_data bts_data;

/* bts_data is built from the input and can be user modified  */
/* except where mentioned                                     */

struct bts_data {
        int n;                 /* nodes are numbered 0,1,...,n */
        int m;                 /* m edges                      */
        int A[MAXN][MAXN];     /* the graph */
        int E[MAXN][MAXN];     /*  E(i,j)= j-th edge adjacent to i or n+1 if done */
        long lexmin[MAXN];      /* lex min spanning tree used for root of reverse search */
        int countonly;         /* TRUE means nodes are counted but not output */
        int prune;             /* mark unexplored only if nchild > prune */
        int debug;             /* TRUE means extra debug output is produced   */
        FILE *output;
        FILE *err;
};


int n;                /* vertices are numbered 0..n  */
int m;                /* number of edges in graph    */

char name[] = "btree";

int  next(tree v, int* r, int* s, bts_data *b); /* get next candidate edge for reverse  */
int ancestor (tree v, int i, int j);            /* return true if i is an ancestor of j */
void lexmin_tree(bts_data *b);                  /* computes lex min spanning tree of input G*/

/*************************************************************************/
/* Required routines for multithreading:                                 */
/*  bts_init, get_root, put_output freebdata bts                         */
/*************************************************************************/

/* bts_init is executed at startup to process command line and build bts_data  */

bts_data *bts_init(int argc, char **argv)
{
        int i, j, k, n,m,N;
        bts_data *b = malloc(sizeof(bts_data));


        b->countonly=FALSE;
        b->debug=FALSE;
        b->prune=-1;           /* no pruning of unexplored nodes */

/* set up I/O handling */
        b->output = stdout;
        b->err = stderr;

/* process command line arguments                     */
/* some error checking may be nice !                  */



        i=1;
        while (i<argc)
          {
          if (strcmp(argv[i],"-countonly") == 0)
             b->countonly=TRUE;
          if (strcmp(argv[i],"-prune") == 0)
             {
             b->prune = atol(argv[++i]);
	     fprintf(b->err, "-prune not yet implemented!");
             }
          if (strcmp(argv[i],"-debug") == 0)
             b->debug=TRUE;
          i++;
          }

/* build bts_data structure from Data *in  */

        i = scanf("%d %d\n", &N, &b->m);
        if (i == EOF )
	{
	   fprintf(b->err, "no input found\n");
           return NULL;
        }

        b->n =N-1;     /* n is max vertex label */
        n=b->n;
        m=b->m;

/* read input data containing the input graph */

        for (i=0; i<=n; i++)
                for (j=0; j<=n; j++)
                        b->A[i][j] = 0;

        for (i=0; i<m; i++)
        {
                scanf("%d %d\n", &j, &k);
                if(j<0 || j>n || k< 0 || k > n  )
                {
		   fprintf(b->err, "bad edge j=%d k=%d: required 0<=j<k<=%d\n",j,k,n);
                   return NULL;
                }
                b->A[k][j] = 1;
                b->A[j][k] = 1;
        }

         fprintf(b->err,"\n");

        lexmin_tree(b);    /* compute lex min spanning tree */

/* build edge lists that exclude treeroot edges */

       for (i=0;i<=n;i++)
       {
           k=0;
           for (j=0;j<=n;j++)
           { if(b->A[i][j] && j!=b->lexmin[i] )
                {
                 b->E[i][k]=j;
                 b->A[i][j]=k;    /* for fast lookup in next(...) */
                 k++;
                }
              b->E[i][k]=n+1;
           }
        }
        b->A[1][1]= -1;                /* for first time i=j=1 case in next */

/* output problem details */

         if(b->debug)
              for (i=0; i<=n; i++)
                 {
                 fprintf(b->err,"\n");
                 for (j=0; j<=n; j++)
                       fprintf(b->err," %d",b->A[i][j]);
                 }
         fprintf(b->err,"\ntreeroot:");
         for (i=1; i<=n; i++)
               fprintf(b->err," %d,%ld",i,b->lexmin[i]);
         fprintf(b->err,"\nremaining graph:");
         for (i=1;i<=n;i++)
           {
             k=0;
              fprintf(b->err,"\n i=%d:",i);
             while (b->E[i][k] != n+1)
                {
                 fprintf(b->err," %d",b->E[i][k]);
                 k++;
                }

            }
              fprintf(b->err,"\n\n");
                                 /* end one time output */


        return b;
}                        /* end of bts_init */


/* the root node of the tree search is built here */


void put_output (bts_data *b, tree v, long depth, long unexplored)
{
 int i;
 if(!b->countonly)
 {
   for (i = 1; i <= b->n; i++)
      fprintf (b->output," %ld", v[i]);
   fprintf (b->output, " d=%ld",depth);
   if(unexplored)
       fprintf (b->output, "unexplored");
   fprintf (b->output, "\n");
 }
}

void freebdata(bts_data *b)
/* modify if other methods to free data are required */
{
        free(b);
}


void lexmin_tree(bts_data *b)
/* build spanning tree with vertex 0 as root */
/* tree is also save in Data *in for faster parallel init */
{
  int i,j,k;
  int n=b->n;
  long *v;

  v=&b->lexmin[0];

  for (i = 1; i <= n; i++)
       v[i]= -1;          /*no parents*/
  v[0]=0;                 /* tree root */
  k=0;                    /* counter for non-root nodes with parents */
  i=0; j=0;
  while (k<n)             /* loops forever if tree disconnected */
     {
       j++;
       if ( i == j) j++;
       if (j > n )
           {  i++;
              if(i>n) i=0;  /* we missed somebody so start again */
              j=0;
           }
      if (b->A[i][j])
        {
         if (v[i]>=0 && v[j]<0)
            {
             k++;
             v[j]=i;
            }
         else 
           if (v[i]<0 && v[j]>=0) /* both negative is possible - hook them up! */
            {
             k++;
             v[i]=j;
            }
         }
      } /* end while k<n */


}  /*  end of lexmin_tree */



void copy (tree w, tree v)
{
  int i;
  for (i = 0; i <= n; i++)
    w[i] = v[i];
}

int equal (tree w, tree v)
{
  int i;
  for (i = 1; i <= n; i++)
    if (w[i] != v[i]) return FALSE;
  return TRUE;
}


int  next(tree v, int* r, int* s, bts_data *b)
/* increment j and possibly i until the next candidate edge ij for reverse is found */
/* conditions are: (a) ij is an edge of graph that is not in tree v or treeroot 
                   (b) i is not an ancestor of j in tree v
                   (c) i v[i] is not an edge of treeroot (otherwise reverse is false)
*/
{
 int i=(*r), j=(*s);
 int k;                   /* E(i,k)=j */
 int p;
 
/* first time i=1,j=1 and A(1,1)=-1; */

      k=b->A[i][j];             
      k++;

 while (i <= n)           /* loop  ends when i=n+1 k=0 */
  {
     if (v[i] == b->lexmin[i]) /* otherwise reverse is false */
       {
        j=b->E[i][k];
	while (j<=n)
           {
             if(j!=v[i] )   /* check ancestor relation */
                  {
                    p=v[j];
                    while ( p != 0 && p != i ) p=v[p];
                    if (p==0)
                      {
                        (*r)=i; (*s)=j;
                        return TRUE;
                      }  
                  }
              k++;
              j=b->E[i][k];
            }
        }
     else
       if (b->lexmin[i] == 0)  /* 0 can never become an ancestor of i */
         i=n;                   /* v is finished */
     i++;
     k=0;
  }
 (*r)=i; (*s)=j;
 return FALSE;
}

int ancestor (tree v, int i, int j)   /* return true if i is an ancestor of j */
                                      /* assumes  vertex 0 is root of tree v  */
{
 int p;
 p=v[j];
 while ( p != 0 && p != i ) p=v[p];
 return (p==i); 
}

void f (tree v,int *i, int *j, tree treeroot)      /* local search function:  v<-f(v) */
                                                /* indices i, j restored for backtrack*/

{
  int k=1;
  while ( k <= n && ( v[k] ==  treeroot[k] || ancestor(v,k,treeroot[k])))
              k++;
  if (k <= n)                                   /* v is not root of tree */
   { (*i)=k; (*j)=v[k];
     v[k]=treeroot[k];
   }
}

int reverse (bts_data *b,tree v, int i, int j)
{       /* next(...) has ensured v[i]=j is a valid edge insertion */
  long savevi;
  int k=1;
  savevi=v[i];
  v[i]=j;         /* we check directly the f function without a call */
  while ( k <= n && ( v[k] ==  b->lexmin[k] || ancestor(v,k,b->lexmin[k])))
              k++;
  v[i]=savevi;
  return (i==k);
}


long bts(bts_data *b, tree v, long max_depth, long max_nodes)
{
  long  depth, *treeroot;
  long  unexplored;
  long init_depth;
  int i=1, j=1;
  int done;
  long count=0;

  treeroot=&b->lexmin[0];

  n = b->n;
  m = b->m;
  depth=0;
  init_depth=depth;   /* depth of root of this subtree */
  if (max_depth > n-1 )
      max_depth =n-1;    /* max depth is n-1, so no need to search these nodes */

  do
    {
      unexplored = FALSE;
      done=FALSE;
      while ( next(v,&i,&j,b) && !done )
      {
         if ( reverse (b,v,i, j))
          {                        /* forward step */
             v[i]=j;       /* equivalent to Adj (v,i,j,b) */               
             depth++;
             count++;
             if (count >= max_nodes || depth == max_depth )
                 {
                   done=TRUE;
                   if (depth < n-1)   /* max depth of this problem is n-1 */  
                        unexplored=TRUE;
                 }
             put_output(b,v,depth,unexplored);
             i=1; j=1;
          }
      }   /* end while */

      if(depth > init_depth)
          {                        /* backtrack step */
             f(v,&i,&j,treeroot);       
             (depth)--;
          }
    } while( depth > init_depth || i <= n);
  return count;
}                 /* end of bts  */


int main(int argc, char **argv) /* generate top sorts */
{
  bts_data *b;
  tree treeroot;
  long count;
  long max_depth;
  long max_nodes;       /* replacing maxcobasis */
  int i;

  fprintf(stderr,"\n*atree 2016.6.20:");

  b=bts_init(argc,argv);
  n=b->n;
  max_depth=b->n-1;
  max_nodes=LONG_MAX;

  for (i=1; i < argc; i++)
    {
          if (strcmp(argv[i],"-maxd") == 0)
                max_depth = atoi(argv[i+1]);
          if (strcmp(argv[i],"-maxnodes") == 0)
                 max_nodes = atoi(argv[i+1]);
    }

  for(i=0;i<=n;i++) treeroot[i]=b->lexmin[i];
  put_output(b,treeroot,0,0);

/* the tree search is now initiated from root */

  count = bts(b,treeroot,max_depth,max_nodes);

  fprintf (stderr,"\nnumber of trees=%ld\n", count+1); /* count the root */

  freebdata(b);
  return 0;
}
