/* sat.c
 * simple distributed SAT solver using MiniSAT-2 and mts.
 * Use -noshare to disable conflict-clause sharing. 
 * Use -norestart to disable restarts
 * Use -nosimp to disable simplification
 * Use -v to get counting stats
 * Use -confbudg to try budgeting -maxnodes as #conflicts, not
 *     #decisions
 * Use -propbudg to try budgeting -maxnodes as #propagations, not
 *     #decisions
 */

#include "mts.h"
#include "minisat.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define tsprintf stream_printf
#define tstream mts_stream
#define TRUE 1
#define FALSE 0

#ifdef SATVERBOSE
#define debugprint(a) printf a
#else
#define debugprint(a)
#endif

struct bts_data {
	minisat_solver *solver;
	tstream *output;
	tstream *err;
	int verbose; /* print stats? (bool) */
	int budget_type; /* 0: decisions 1: conflicts 2: propagations */
};

char name[] = "mtsat";

Node *assumptions;
int rank;
shared_data *mysdata;
int mtsat_share; /* share conflict clauses? currently bool, units only */

const mtsoption bts_options[] = {
 {"-noshare", 0},
 {"-norestart", 0},
 {"-nosimp", 0},
 {"-confbudg", 0},
 {"-propbudg", 0},
 {"-v", 0}
};
const int bts_options_size = 6;

void proc_conf(const shared_data *, bts_data *, int);

void bts_finish(bts_data *b, shared_data *sdata, int size, int checkpointing)
{
	int i;
	int sat=0;
	unsigned long counting[4]={0};
	open_outputblock();
	for (i=0; i<size && sat==0; i++)
	{
		if (b->verbose && sdata[i].size_sdlong != 0)
		{
			counting[0] += sdata[i].sdlong[0];
			counting[1] += sdata[i].sdlong[1];
			counting[2] += sdata[i].sdlong[2];
			counting[3] += sdata[i].sdlong[3];
		}
		if (sdata[i].size_sdchar==0)
			continue;
		tsprintf(b->output, "%s", sdata[i].sdchar);
		sat=1;
	}
	if (sat==0 && checkpointing==0)
		tsprintf(b->output, "\nUNSATISFIABLE\n\n");
	close_outputblock();

	flush_stream(b->output);

	if (b->verbose)
		/* print counting stats */
		tsprintf(b->err, "restarts\t: %lu\nconflicts\t: %lu\ndecisions\t: %lu\npropagations\t: %lu\n",
		 	counting[2], counting[0], counting[1], counting[3]);
	return;
}

bts_data *bts_init(int argc, char **argv, Data *in, int proc_no)
{
	int i, nv, simp=1;
	bts_data *b = malloc(sizeof(bts_data));

	b->output = open_stream(MTSOUT);
	b->err = open_stream(MTSERR);
	b->solver = minisat_new();
	mtsat_share = 1; /* default: share unit conflicts */
	b->verbose = 0; /* default: don't collect/print stats */
	b->budget_type = 0; /* default: budget is #decisions */
	parse_dimacs(in->input, b->solver);
	debugprint(("%d: num clauses: %d vars: %d freevars: %d\n",
		   proc_no, minisat_num_clauses(b->solver),
		   minisat_num_vars(b->solver),
		   minisat_num_freeVars(b->solver)));

	for (i=1; i<argc; i++)
	{
		if (!strcmp(argv[i], "-noshare"))
		{
			tsprintf(b->err, "* clause sharing disabled\n");
			mtsat_share = 0;
			continue;
		}
		if (!strcmp(argv[i], "-norestart"))
		{
			minisat_set_norestart(b->solver);
			tsprintf(b->err, "* restarts disabled\n");
			continue;
		}
		if (!strcmp(argv[i], "-nosimp"))
		{
			simp = 0;
			tsprintf(b->err, "* not simplifying\n");
			continue;
		}
		if (!strcmp(argv[i], "-v"))
		{
			b->verbose = 1;
			continue;
		}
		if (!strcmp(argv[i], "-confbudg"))
		{
			b->budget_type = 1;
			tsprintf(b->err, "* conflict budget\n");
			continue;
		}
		if (!strcmp(argv[i], "-propbudg"))
		{
			b->budget_type = 2;
			tsprintf(b->err, "* propagation budget\n");
			continue;
		}
	}

	if (simp)
	{
		debugprint(("%d: simplifying\n", proc_no));
		/* minisat_simplify(b->solver); */
		minisat_eliminate(b->solver, minisat_l_True);
		debugprint(("%d: done simplifying\n", proc_no));
		debugprint(("%d: num clauses: %d vars: %d freevars: %d\n",
                   proc_no, minisat_num_clauses(b->solver),
                   minisat_num_vars(b->solver),
                   minisat_num_freeVars(b->solver)));
	}


	nv = minisat_num_vars(b->solver);
#if 1
	for (i=0; i<nv; i++) /* freeze variables that may be used as assumptions :-( */
		minisat_setFrozen(b->solver, i, minisat_l_True);

#endif
	return b;
}

Node *get_root(bts_data *b)
{
	Node *root = malloc(sizeof(Node));
	root->vlong = NULL;
	root->size_vlong = 0;
	root->vint = NULL;
	root->size_vint = 0;
	root->vchar = NULL;
	root->size_vchar = 0;
	root->vfloat = NULL;
	root->size_vfloat = 0;
	root->depth = 0;
	root->unexplored = FALSE;
	return root;
}

void freebdata(bts_data *b)
{
	minisat_delete(b->solver);
	free(b);
	return;
}

/* solve the SAT instance in b with assumptions in assumps,
 * with budget (max_depth, max_nodes)
 */
long bts(bts_data *b, Node *assumps, long max_depth, long max_nodes,
	 shared_data *sdata, int rk, int size)
{
	int i;
	int len;
	long nodecount;
	long oldcount;

	minisat_bool ret;
	minisat_Lit *a = malloc(sizeof(minisat_Lit)*assumps->size_vint);

	rank = rk;
	assumptions = assumps;

	for (i=0; i<size; i++)
	{
		if (sdata[i].modified == 1 && sdata[i].size_sdint > 0)
			proc_conf(sdata+i, b, rank);
	}

	mysdata = sdata+rank;
	if (mysdata->sdint == NULL)
	{
		mysdata->sdint = malloc(sizeof(int)*32);
		mysdata->size_sdint = 32;
	}
	mysdata->sdint[0] = 0; /* no new unit conflicts yet */

	for (i=0; i<assumps->size_vint; i++)
		a[i] = assumps->vint[i];

	if (b->budget_type == 0)
	{
		minisat_set_dec_budget(b->solver, max_nodes);
		oldcount = minisat_num_decisions(b->solver);
	}
	else if (b->budget_type == 1)
	{
		minisat_set_conf_budget(b->solver, max_nodes);
		oldcount = minisat_num_conflicts(b->solver);
	}
	else
	{
		minisat_set_prop_budget(b->solver, max_nodes);
		oldcount = minisat_num_propagations(b->solver);
	}
	debugprint(("\n%d: Solving w budg %ld and assumptions: ", rank,
		    max_nodes));
	for (i=0; i<assumps->size_vint; i++)
		debugprint(("%d ", assumps->vint[i]));
	debugprint(("\n\n"));

	ret = minisat_limited_solve(b->solver, assumps->size_vint, a);
	free(a);

	if (b->budget_type == 0)
                nodecount = minisat_num_decisions(b->solver);
        else if (b->budget_type == 1)
                nodecount = minisat_num_conflicts(b->solver);
        else
                nodecount = minisat_num_propagations(b->solver);

	nodecount -= oldcount;

	/* update counting stats */
	/* only needed for outputting stats in bts_finish,
	 * remove to decrease use of sdata
	 */
	if (b->verbose)
	{
		if (sdata[rank].size_sdlong == 0)
		{
			sdata[rank].sdlong = malloc(sizeof(long)*4);
			sdata[rank].sdlong[0] = sdata[rank].sdlong[1] = 0;
			sdata[rank].sdlong[2] = sdata[rank].sdlong[3] = 0;
			sdata[rank].size_sdlong = 4;
		}
		sdata[rank].sdlong[0] = minisat_num_conflicts(b->solver);
		sdata[rank].sdlong[1] = minisat_num_decisions(b->solver);
		sdata[rank].sdlong[2] = minisat_num_restarts(b->solver);
		sdata[rank].sdlong[3] = minisat_num_propagations(b->solver);
		sdata[rank].modified = 1;
	}
	if (ret == minisat_l_True) /* found satisfying assignment */
	{
		cleanstop(0);
		len = 0;
		len += snprintf(NULL, 0, "\nSAT\n");
		for (i=0; i<minisat_num_vars(b->solver); i++)
		{
			if (minisat_modelValue_Var(b->solver, i) ==
			    minisat_l_True)
				len+=snprintf(NULL, 0, "%d ", i+1);
			else
				len+=snprintf(NULL, 0, "-%d ", i+1);
		}
		len += snprintf(NULL, 0, "\n\n");
		sdata[rank].sdchar = malloc(sizeof(char)*(len+1));
		sdata[rank].size_sdchar = len+1;
		len = 0;
		len += sprintf(sdata[rank].sdchar, "\nSAT\n");
		for (i=0; i<minisat_num_vars(b->solver); i++)
		{
			if (minisat_modelValue_Var(b->solver, i) ==
			    minisat_l_True)
				len+=sprintf(sdata[rank].sdchar+len,"%d ",i+1);
			else
				len+=sprintf(sdata[rank].sdchar+len,"-%d ",i+1);
		}
		len += sprintf(sdata[rank].sdchar+len, "\n\n");
		sdata[rank].modified = 1;
		return nodecount;
	}
	else if (ret == minisat_l_False) /* unsat under these assumptions */
	{
		debugprint(("%d: subproblem UNSAT\n", rank));
	 	return nodecount;
	}

	return nodecount;
}

/* out of budget, blit[0]...blit[numlit-1] are the decision literals
 * we'll backtrack on - split subproblems and return unexplored mts style
 * i.e., neg(blit[0]) added to assumps and {blit[0], neg(blit[1]} added, etc.
 */
void mtsat_backtrack(int *blit, int numlit)
{
	int i, sa = assumptions->size_vint;
	debugprint(("%d: subproblem undet, backtrack %d decision vars\n", rank,
		    numlit));
	assumptions->vint = realloc(assumptions->vint, sizeof(int)*(assumptions->size_vint+numlit));
	for (i=0; i<numlit; i++)
		assumptions->vint[sa+i] = blit[i];
	for (i=0; i<numlit; i++)
	{
		assumptions->vint[sa+i] = minisat_negate(assumptions->vint[sa+i]);
		assumptions->size_vint++;
		return_unexp(assumptions);
		assumptions->vint[sa+i] = minisat_negate(assumptions->vint[sa+i]);
	}
	/* and continue along the path that we were on */
	assumptions->vint[numlit-1] = blit[numlit-1];
	return_unexp(assumptions);
	return;
}

/* got a new learnt unit clause */
void mtsat_newlearnt(int lit)
{
	int num = mysdata->sdint[0];
	int *tmp;

	if (!mtsat_share) /* don't share conflict clauses? */
		return;

	debugprint(("  %d: sharing new unit %d\n", rank, lit)); 
	if (mysdata->size_sdint <= (num+1))
	{
		tmp = realloc(mysdata->sdint, sizeof(int)*(num+2));
		if (tmp==NULL)
			return; /* don't share this one */
		mysdata->sdint = tmp;
		mysdata->size_sdint = num+2;
	}
	mysdata->sdint[0]++;
	mysdata->sdint[num+1] = lit;
	mysdata->modified = 1;
}

/* sdata has a conflict clause produced by share_conflict.
 * Add it to the solver
 */

void proc_conf(const shared_data *sdata, bts_data *b, int rank)
{
	minisat_solver *s = b->solver;
	int i=0;

	for (i=1; i<=sdata->sdint[0]; i++)
	{
		debugprint((" %d: adding conflict clause: %d\n", rank,
			    sdata->sdint[i]));
		minisat_addClause_begin(s);
		minisat_addClause_addLit(s, sdata->sdint[i]);
		minisat_addClause_commit(s);
	}
}
